from json import loads, dumps
from typing import List


class Transaction:
    def __init__(self, size: int, block_number: int, position_in_the_block: int, fee: int, extraFee: int, type: str,
                 sender: str, receiver: str, nonce: int, hash: str, timestamp: int, value: int, raw_transaction: bytes,
                 chainId: bytes, hasError: bool):
        self._size = size
        self._block_number = block_number
        self._position_in_the_block = position_in_the_block
        self._fee = fee
        self._extraFee = extraFee
        self._type = type
        self._sender = sender
        self._receiver = receiver
        self._nonce = nonce
        self._hash = hash
        self._timestamp = timestamp
        self._value = value
        self._raw_transaction = raw_transaction
        self._chainId = chainId
        self._hasError = hasError

    @classmethod
    def from_json(cls, json_data, block_number, timestamp, position_in_the_block):
        size = json_data.get('size', 0)
        block_number = block_number
        position_in_the_block = position_in_the_block
        fee = json_data.get('fee', 0)
        extraFee = json_data.get('extraFee', 0)  # Ensure this field is included
        _type = json_data.get('type', 'unknown')
        sender = json_data.get('sender', '0x')
        receiver = json_data.get('receiver', '0x')
        nonce = json_data.get('nonce', 0)
        _hash = json_data.get('hash', '0x')
        timestamp = timestamp
        value = json_data.get('value', 0)
        raw_transaction = bytes.fromhex(json_data.get('rawTransaction', ''))
        chainId = int.to_bytes(json_data.get('chainId', 0), 1, 'big')
        hasError = not (json_data.get("success", True))  # Fix typo from "sucess" to "success"

        return cls(size=size, block_number=block_number, position_in_the_block=position_in_the_block, fee=fee,
                   extraFee=extraFee, type=_type, sender=sender, receiver=receiver, nonce=nonce, hash=_hash,
                   timestamp=timestamp, value=value, raw_transaction=raw_transaction, chainId=chainId,
                   hasError=hasError)

    @classmethod
    def transaction_from_json(cls, json_data, block_number, timestamp, position_in_the_block):
        txn_type = json_data.get('type', 'Unknown').lower()
        if txn_type == ClaimSpotTransaction.TYPE.lower():
            return ClaimSpotTransaction.from_json(json_data, block_number, timestamp, position_in_the_block)
        elif txn_type == TransferTransaction.TYPE.lower():
            return TransferTransaction.from_json(json_data, block_number, timestamp, position_in_the_block)
        elif txn_type == VidaDataTransaction.TYPE.lower():
            return VidaDataTransaction.from_json(json_data, block_number, timestamp, position_in_the_block)
        elif txn_type == DelegateTransaction.TYPE.lower():
            return DelegateTransaction.from_json(json_data, block_number, timestamp, position_in_the_block)
        elif txn_type == WithdrawTransaction.TYPE.lower():
            return WithdrawTransaction.from_json(json_data, block_number, timestamp, position_in_the_block)
        elif txn_type == JoinTransaction.TYPE.lower():
            return JoinTransaction.from_json(json_data, block_number, timestamp, position_in_the_block)
        elif txn_type == ClaimVidaIdTransaction.TYPE.lower():
            return ClaimVidaIdTransaction.from_json(json_data, block_number, timestamp, position_in_the_block)
        elif txn_type == SetGuardianTransaction.TYPE.lower():
            return SetGuardianTransaction.from_json(json_data, block_number, timestamp, position_in_the_block)
        elif txn_type == PayableVidaDataTransaction.TYPE.lower():
            return PayableVidaDataTransaction.from_json(json_data, block_number, timestamp, position_in_the_block)
        elif txn_type == GuardianApprovalTransaction.TYPE.lower():
            return GuardianApprovalTransaction.from_json(json_data, block_number, timestamp, position_in_the_block)
        elif txn_type == ConduitApprovalTransaction.TYPE.lower():
            return ConduitApprovalTransaction.from_json(json_data, block_number, timestamp, position_in_the_block)
        elif txn_type == RemoveGuardianTransaction.TYPE.lower():
            return RemoveGuardianTransaction.from_json(json_data, block_number, timestamp, position_in_the_block)
        elif txn_type == SetConduitsTransaction.TYPE.lower():
            return SetConduitsTransaction.from_json(json_data, block_number, timestamp, position_in_the_block)
        elif txn_type == AddConduitsTransaction.TYPE.lower():
            return AddConduitsTransaction.from_json(json_data, block_number, timestamp, position_in_the_block)
        elif txn_type == MoveStakeTransaction.TYPE.lower():
            return MoveStakeTransaction.from_json(json_data, block_number, timestamp, position_in_the_block)
        elif txn_type == ChangeEarlyWithdrawPenaltyProposalTransaction.TYPE.lower():
            return ChangeEarlyWithdrawPenaltyProposalTransaction.from_json(json_data, block_number, timestamp,
                                                                           position_in_the_block)
        elif txn_type == ChangeFeePerByteProposalTransaction.TYPE.lower():
            return ChangeFeePerByteProposalTransaction.from_json(json_data, block_number, timestamp,
                                                                 position_in_the_block)
        elif txn_type == ChangeMaxTxnSizeProposalTransaction.TYPE.lower():
            return ChangeMaxTxnSizeProposalTransaction.from_json(json_data, block_number, timestamp,
                                                                 position_in_the_block)
        elif txn_type == ChangeMaxBlockSizeProposalTransaction.TYPE.lower():
            return ChangeMaxBlockSizeProposalTransaction.from_json(json_data, block_number, timestamp,
                                                                   position_in_the_block)
        elif txn_type == ChangeOverallBurnPercentageProposalTransaction.TYPE.lower():
            return ChangeOverallBurnPercentageProposalTransaction.from_json(json_data, block_number, timestamp,
                                                                            position_in_the_block)
        elif txn_type == ChangeRewardPerYearProposalTransaction.TYPE.lower():
            return ChangeRewardPerYearProposalTransaction.from_json(json_data, block_number, timestamp,
                                                                    position_in_the_block)
        elif txn_type == ChangeValidatorCountLimitProposalTransaction.TYPE.lower():
            return ChangeValidatorCountLimitProposalTransaction.from_json(json_data, block_number, timestamp,
                                                                          position_in_the_block)
        elif txn_type == ChangeValidatorJoiningFeeProposalTransaction.TYPE.lower():
            return ChangeValidatorJoiningFeeProposalTransaction.from_json(json_data, block_number, timestamp,
                                                                          position_in_the_block)
        elif txn_type == ChangeVidaIdClaimingFeeProposalTransaction.TYPE.lower():
            return ChangeVidaIdClaimingFeeProposalTransaction.from_json(json_data, block_number, timestamp,
                                                                      position_in_the_block)
        elif txn_type == ChangeVidaOwnerTxnFeeShareProposalTransaction.TYPE.lower():
            return ChangeVidaOwnerTxnFeeShareProposalTransaction.from_json(json_data, block_number, timestamp,
                                                                         position_in_the_block)
        elif txn_type == OtherProposalTransaction.TYPE.lower():
            return OtherProposalTransaction.from_json(json_data, block_number, timestamp, position_in_the_block)
        elif txn_type == VoteOnProposalTransaction.TYPE.lower():
            return VoteOnProposalTransaction.from_json(json_data, block_number, timestamp, position_in_the_block)
        else:
            return Transaction.from_json(json_data, block_number, timestamp, position_in_the_block)

    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, value):
        self._size = value

    @property
    def position_in_the_block(self):
        return self._position_in_the_block

    @position_in_the_block.setter
    def position_in_the_block(self, value):
        self._position_in_the_block = value

    @property
    def fee(self):
        return self._fee

    @fee.setter
    def fee(self, value):
        self._fee = value

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, value):
        self._type = value

    @property
    def sender(self):
        return self._sender

    @sender.setter
    def sender(self, value):
        self._sender = value

    @property
    def receiver(self):
        return self._receiver

    @receiver.setter
    def receiver(self, value):
        self._receiver = value

    @property
    def nonce(self):
        return self._nonce

    @nonce.setter
    def nonce(self, value):
        self._nonce = value

    @property
    def hash(self):
        return self._hash

    @hash.setter
    def hash(self, value):
        self._hash = value

    @property
    def block_number(self):
        return self._block_number

    @block_number.setter
    def block_number(self, value):
        self._block_number = value

    @property
    def timestamp(self):
        return self._timestamp

    @timestamp.setter
    def timestamp(self, value):
        self._timestamp = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value

    @property
    def raw_transaction(self):
        return self._raw_transaction

    @raw_transaction.setter
    def raw_transaction(self, value):
        self._raw_transaction = value

    def to_json(self):
        json_object = {
            "size": self.size,
            "positionInTheBlock": self.position_in_the_block,
            "fee": self.fee,
            "type": self.type,
            "sender": self.sender,
            "receiver": self.receiver,
            "nonce": self.nonce,
            "hash": self.hash,
            "blockNumber": self.block_number,
            "timestamp": self.timestamp,
            "value": self.value,
            "rawTransaction": self.raw_transaction.hex()
        }
        return json_object


class ClaimSpotTransaction(Transaction):
    TYPE = "Validator Claim Spot"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._validator: str = ""

    @classmethod
    def from_json(cls, json_data, block_number, timestamp, position_in_the_block):
        instance = super().from_json(json_data, block_number, timestamp, position_in_the_block)
        instance._validator = instance.sender
        return instance

    @property
    def validator(self) -> str:
        return self._validator

    @validator.setter
    def validator(self, value: str):
        self._validator = value

    def to_json(self):
        json_object = super().to_json()
        json_object['validator'] = self.validator
        return json_object


class ClaimVidaIdTransaction(Transaction):
    TYPE = "Claim VIDA ID"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._vida_id: int = 0

    @classmethod
    def from_json(cls, json_data, block_number, timestamp, position_in_the_block):
        instance = super().from_json(json_data, block_number, timestamp, position_in_the_block)
        instance._vida_id = json_data.get('vidaId', 0)
        return instance

    @property
    def vida_id(self) -> int:
        return self._vida_id

    def to_json(self):
        json_object = super().to_json()
        json_object['vidaId'] = self.vida_id
        return json_object


class ConduitApprovalTransaction(Transaction):
    TYPE = "Conduit Approval"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._vida_id: int = 0
        self._transactions: List[Transaction] = []

    @classmethod
    def from_json(cls, json_data, block_number, timestamp, position_in_the_block):
        instance = super().from_json(json_data, block_number, timestamp, position_in_the_block)
        instance._vida_id = json_data.get('vidaId', 0)
        transactions_data = json_data.get('transactions', [])
        instance._transactions = [
            Transaction.from_json(loads(transaction_data), block_number, timestamp, position_in_the_block)
            for transaction_data in transactions_data
        ]
        return instance

    @property
    def vida_id(self) -> int:
        return self._vida_id

    @property
    def transactions(self) -> List[Transaction]:
        return self._transactions

    def to_json(self):
        json_object = super().to_json()
        json_object['vidaId'] = self.vida_id
        json_object['transactions'] = [transaction.to_json() for transaction in self.transactions]
        return json_object


class DelegateTransaction(Transaction):
    TYPE = "Delegate"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._validator: str = ""

    @classmethod
    def from_json(cls, json_data, block_number, timestamp, position_in_the_block):
        instance = super().from_json(json_data, block_number, timestamp, position_in_the_block)
        instance._validator = json_data.get('validator', '0x')
        return instance

    @property
    def validator(self) -> str:
        return self._validator

    def to_json(self):
        json_object = super().to_json()
        json_object['validator'] = self.validator
        return json_object


class GuardianApprovalTransaction(Transaction):
    TYPE = "Guardian Approval"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._transactions: List[Transaction] = []

    @classmethod
    def from_json(cls, json_data, block_number, timestamp, position_in_the_block):
        instance = super().from_json(json_data, block_number, timestamp, position_in_the_block)
        transactions_data = json_data.get('transactions', [])
        instance._transactions = [
            Transaction.from_json(transaction_data, block_number, timestamp, position_in_the_block)
            for transaction_data in transactions_data
        ]
        return instance

    @property
    def transactions(self) -> List[Transaction]:
        return self._transactions

    def to_json(self):
        json_object = super().to_json()
        json_object['transactions'] = [transaction.to_json() for transaction in self.transactions]
        return json_object


class JoinTransaction(Transaction):
    TYPE = "Validator Join"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._validator: str = ""
        self._ip: str = ""

    @classmethod
    def from_json(cls, json_data, block_number, timestamp, position_in_the_block):
        instance = super().from_json(json_data, block_number, timestamp, position_in_the_block)
        instance._validator = json_data.get('sender', '0x')
        instance._ip = json_data.get('ip', '')
        return instance

    @property
    def validator(self) -> str:
        return self._validator

    @property
    def ip(self) -> str:
        return self._ip

    def to_json(self):
        json_object = super().to_json()
        json_object['validator'] = self.validator
        json_object['ip'] = self.ip
        return json_object


class PayableVidaDataTransaction(Transaction):
    TYPE = "Payable VIDA Data"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._vida_id: int = 0
        self._data: str = "0x"

    @classmethod
    def from_json(cls, json_data, block_number, timestamp, position_in_the_block):
        instance = super().from_json(json_data, block_number, timestamp, position_in_the_block)
        instance._vida_id = json_data.get('vidaId', 0)
        instance._data = json_data.get('data', '0x')
        return instance

    @property
    def vida_id(self) -> int:
        return self._vida_id

    @property
    def data(self) -> str:
        return self._data

    def to_json(self):
        json_object = super().to_json()
        json_object['vidaId'] = self.vida_id
        json_object['data'] = self.data
        return json_object


class RemoveGuardianTransaction(Transaction):
    TYPE = "Remove Guardian"

    @classmethod
    def from_json(cls, json_data, block_number, timestamp, position_in_the_block):
        instance = super().from_json(json_data, block_number, timestamp, position_in_the_block)
        return instance


class SetGuardianTransaction(Transaction):
    TYPE = "Set Guardian"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._guardian: str = "0x"
        self._expiry_date: int = 0

    @classmethod
    def from_json(cls, json_data, block_number, timestamp, position_in_the_block):
        instance = super().from_json(json_data, block_number, timestamp, position_in_the_block)
        instance._guardian = json_data.get('guardian', '0x')
        instance._expiry_date = json_data.get('expiryDate', 0)
        return instance

    @property
    def guardian(self) -> str:
        return self._guardian

    @property
    def expiry_date(self) -> int:
        return self._expiry_date

    def to_json(self):
        json_object = super().to_json()
        json_object['guardian'] = self.guardian
        json_object['expiryDate'] = self.expiry_date
        return json_object


class TransferTransaction(Transaction):
    TYPE = "Transfer"

    def __init__(self, *args, chainId: bytes = b'', **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def from_json(cls, json_data, block_number, timestamp, position_in_the_block):
        return super().from_json(json_data, block_number, timestamp, position_in_the_block)


class VidaDataTransaction(Transaction):
    TYPE = "VIDA Data"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._vida_id: int = 0
        self._data: str = ""

    @classmethod
    def from_json(cls, json_data, block_number, timestamp, position_in_the_block):
        instance = super().from_json(json_data, block_number, timestamp, position_in_the_block)
        instance._vida_id = json_data.get('vidaId', 0)
        instance._data = json_data.get('data', "")
        return instance

    @property
    def vida_id(self) -> int:
        return self._vida_id

    @property
    def data(self) -> str:
        return self._data

    def to_json(self):
        json_object = super().to_json()
        json_object['vidaId'] = self.vida_id
        json_object['data'] = self.data
        json_object['type'] = "VIDA Data"
        return json_object


class WithdrawTransaction(Transaction):
    TYPE = "Withdraw"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._validator: str = "0x"
        self._shares: int = 0

    @classmethod
    def from_json(cls, json_data, block_number, timestamp, position_in_the_block):
        instance = super().from_json(json_data, block_number, timestamp, position_in_the_block)
        instance._validator = json_data.get('validator', '0x')
        instance._shares = json_data.get('shares', 0)
        return instance

    @property
    def validator(self) -> str:
        return self._validator

    @property
    def shares(self) -> int:
        return self._shares

    def to_json(self):
        json_object = super().to_json()
        json_object['validator'] = self.validator
        json_object['shares'] = self.shares
        return json_object


class AddConduitsTransaction(Transaction):
    TYPE = "Add Conduits"

    def __init__(self, json_data: dict, block_number: int, timestamp: int, position_in_the_block: int,
                 vida_id: int = 0, conduits: List[str] = None):
        super().__init__(json_data, block_number, timestamp, position_in_the_block)
        self.vida_id = json_data.get("vidaId", 0)
        self.conduits = json_data.get("conduits", [])

    def to_json(self) -> dict:
        transaction = super().to_json()
        transaction["vidaId"] = self.vida_id
        transaction["conduits"] = self.conduits
        return transaction


class MoveStakeTransaction(Transaction):
    TYPE = "Move Stake"

    def __init__(self, json_data: dict, block_number: int, timestamp: int, position_in_the_block: int,
                 from_validator: str = "0x", to_validator: str = "0x", shares_amount: int = 0):
        super().__init__(json_data, block_number, timestamp, position_in_the_block)
        self.from_validator = json_data.get("fromValidator", "0x")
        self.to_validator = json_data.get("toValidator", "0x")
        self.shares_amount = json_data.get("sharesAmount", 0)

    def to_json(self) -> dict:
        transaction = super().to_json()
        transaction["fromValidator"] = self.from_validator
        transaction["toValidator"] = self.to_validator
        transaction["sharesAmount"] = self.shares_amount
        return transaction


class RemoveConduitsTransaction(Transaction):
    TYPE = "Remove Conduits"

    def __init__(self, json_data: dict, block_number: int, timestamp: int, position_in_the_block: int,
                 vida_id: int = 0, conduits: List[str] = None):
        super().__init__(json_data, block_number, timestamp, position_in_the_block)
        self.vida_id = json_data.get("vidaId", 0)
        self.conduits = json_data.get("conduits", [])

    def to_json(self) -> dict:
        transaction = super().to_json()
        transaction["vidaId"] = self.vida_id
        transaction["conduits"] = self.conduits
        return transaction


class SetConduitsTransaction(Transaction):
    TYPE = "Set Conduits"

    def __init__(self, json_data: dict, block_number: int, timestamp: int, position_in_the_block: int,
                 vida_id: int = 0, conduits: List[str] = None):
        super().__init__(json_data, block_number, timestamp, position_in_the_block)
        self.vida_id = json_data.get("vidaId", 0)
        self.conduits = json_data.get("conduits", [])

    def to_json(self) -> dict:
        transaction = super().to_json()
        transaction["vidaId"] = self.vida_id
        transaction["conduits"] = self.conduits
        return transaction


class ChangeEarlyWithdrawPenaltyProposalTransaction(Transaction):
    TYPE = "Change Early Withdraw Penalty Proposal"

    def __init__(self, json_data: dict, block_number: int, timestamp: int, position_in_the_block: int,
                 withdrawal_penalty_time: int = 0, withdrawal_penalty: int = 0, title: str = "", description: str = ""):
        super().__init__(json_data, block_number, timestamp, position_in_the_block)
        self.withdrawal_penalty_time = json_data.get("earlyWithdrawTime", 0)
        self.withdrawal_penalty = json_data.get("earlyWithdrawPenalty", 0)
        self.title = json_data.get("title", "")
        self.description = json_data.get("description", "")

    @classmethod
    def from_json(cls, json_data, block_number, timestamp, position_in_the_block):
        instance = super().from_json(json_data, block_number, timestamp, position_in_the_block)
        instance.withdrawal_penalty_time = json_data.get("earlyWithdrawTime", 0)
        instance.withdrawal_penalty = json_data.get("earlyWithdrawPenalty", 0)
        instance.title = json_data.get("title", "")
        instance.description = json_data.get("description", "")
        return instance

    def to_json(self) -> dict:
        transaction = super().to_json()
        transaction["withdrawalPenaltyTime"] = self.withdrawal_penalty_time
        transaction["withdrawalPenalty"] = self.withdrawal_penalty
        transaction["title"] = self.title
        transaction["description"] = self.description
        return transaction


class ChangeFeePerByteProposalTransaction(Transaction):
    TYPE = "Change Fee Per Byte Proposal"

    def __init__(self, json_data: dict, block_number: int, timestamp: int, position_in_the_block: int,
                 fee_per_byte: int = 8, title: str = "", description: str = ""):
        super().__init__(json_data, block_number, timestamp, position_in_the_block)
        self.fee_per_byte = json_data.get("feePerByte", 8)
        self.title = json_data.get("title", "")
        self.description = json_data.get("description", "")

    @classmethod
    def from_json(cls, json_data, block_number, timestamp, position_in_the_block):
        instance = super().from_json(json_data, block_number, timestamp, position_in_the_block)
        instance.fee_per_byte = json_data.get("feePerByte", 8)
        instance.title = json_data.get("title", "")
        instance.description = json_data.get("description", "")
        return instance

    def to_json(self) -> dict:
        transaction = super().to_json()
        transaction["feePerByte"] = self.fee_per_byte
        transaction["title"] = self.title
        transaction["description"] = self.description
        return transaction


class ChangeMaxBlockSizeProposalTransaction(Transaction):
    TYPE = "Change Max Block Size Proposal"

    def __init__(self, json_data: dict, block_number: int, timestamp: int, position_in_the_block: int,
                 max_block_size: int = 4, title: str = "", description: str = ""):
        super().__init__(json_data, block_number, timestamp, position_in_the_block)
        self.max_block_size = json_data.get("maxBlockSize", 4)
        self.title = json_data.get("title", "")
        self.description = json_data.get("description", "")

    @classmethod
    def from_json(cls, json_data, block_number, timestamp, position_in_the_block):
        instance = super().from_json(json_data, block_number, timestamp, position_in_the_block)
        instance.max_block_size = json_data.get("maxBlockSize", 4)
        instance.title = json_data.get("title", "")
        instance.description = json_data.get("description", "")
        return instance

    def to_json(self) -> dict:
        transaction = super().to_json()
        transaction["maxBlockSize"] = self.max_block_size
        transaction["title"] = self.title
        transaction["description"] = self.description
        return transaction


class ChangeMaxTxnSizeProposalTransaction(Transaction):
    TYPE = "Change Max Txn Size Proposal"

    def __init__(self, json_data: dict, block_number: int, timestamp: int, position_in_the_block: int,
                 max_txn_size: int = 8, title: str = "", description: str = ""):
        super().__init__(json_data, block_number, timestamp, position_in_the_block)
        self.max_txn_size = json_data.get("maxTxnSize", 8)
        self.title = json_data.get("title", "")
        self.description = json_data.get("description", "")

    @classmethod
    def from_json(cls, json_data, block_number, timestamp, position_in_the_block):
        instance = super().from_json(json_data, block_number, timestamp, position_in_the_block)
        instance.max_txn_size = json_data.get("maxTxnSize", 8)
        instance.title = json_data.get("title", "")
        instance.description = json_data.get("description", "")
        return instance

    def to_json(self) -> dict:
        transaction = super().to_json()
        transaction["maxTxnSize"] = self.max_txn_size
        transaction["title"] = self.title
        transaction["description"] = self.description
        return transaction


class ChangeOverallBurnPercentageProposalTransaction(Transaction):
    TYPE = "Change Overall Burn Percentage Proposal"

    def __init__(self, json_data: dict, block_number: int, timestamp: int, position_in_the_block: int,
                 burn_percentage: int = 4, title: str = "", description: str = ""):
        super().__init__(json_data, block_number, timestamp, position_in_the_block)
        self.burn_percentage = json_data.get("overallBurnPercentage", 4)
        self.title = json_data.get("title", "")
        self.description = json_data.get("description", "")

    @classmethod
    def from_json(cls, json_data, block_number, timestamp, position_in_the_block):
        instance = super().from_json(json_data, block_number, timestamp, position_in_the_block)
        instance.burn_percentage = json_data.get("overallBurnPercentage", 4)
        instance.title = json_data.get("title", "")
        instance.description = json_data.get("description", "")
        return instance

    def to_json(self) -> dict:
        transaction = super().to_json()
        transaction["burnPercentage"] = self.burn_percentage
        transaction["title"] = self.title
        transaction["description"] = self.description
        return transaction


class ChangeRewardPerYearProposalTransaction(Transaction):
    TYPE = "Change Reward Per Year Proposal"

    def __init__(self, json_data: dict, block_number: int, timestamp: int, position_in_the_block: int,
                 reward_per_year: int = 8, title: str = "", description: str = ""):
        super().__init__(json_data, block_number, timestamp, position_in_the_block)
        self.reward_per_year = json_data.get("rewardPerYear", 8)
        self.title = json_data.get("title", "")
        self.description = json_data.get("description", "")

    @classmethod
    def from_json(cls, json_data, block_number, timestamp, position_in_the_block):
        instance = super().from_json(json_data, block_number, timestamp, position_in_the_block)
        instance.reward_per_year = json_data.get("rewardPerYear", 8)
        instance.title = json_data.get("title", "")
        instance.description = json_data.get("description", "")
        return instance

    def to_json(self) -> dict:
        transaction = super().to_json()
        transaction["rewardPerYear"] = self.reward_per_year
        transaction["title"] = self.title
        transaction["description"] = self.description
        return transaction


class ChangeValidatorCountLimitProposalTransaction(Transaction):
    TYPE = "Change Validator Count Limit Proposal"

    def __init__(self, json_data: dict, block_number: int, timestamp: int, position_in_the_block: int,
                 validator_count_limit: int = 4, title: str = "", description: str = ""):
        super().__init__(json_data, block_number, timestamp, position_in_the_block)
        self.validator_count_limit = json_data.get("validatorCountLimit", 4)
        self.title = json_data.get("title", "")
        self.description = json_data.get("description", "")

    @classmethod
    def from_json(cls, json_data, block_number, timestamp, position_in_the_block):
        instance = super().from_json(json_data, block_number, timestamp, position_in_the_block)
        instance.validator_count_limit = json_data.get("validatorCountLimit", 4)
        instance.title = json_data.get("title", "")
        instance.description = json_data.get("description", "")
        return instance

    def to_json(self) -> dict:
        transaction = super().to_json()
        transaction["validatorCountLimit"] = self.validator_count_limit
        transaction["title"] = self.title
        transaction["description"] = self.description
        return transaction


class ChangeValidatorJoiningFeeProposalTransaction(Transaction):
    TYPE = "Change Validator Joining Fee Proposal"

    def __init__(self, json_data: dict, block_number: int, timestamp: int, position_in_the_block: int,
                 joining_fee: int = 8, title: str = "", description: str = ""):
        super().__init__(json_data, block_number, timestamp, position_in_the_block)
        self.joining_fee = json_data.get("validatorJoiningFee", 8)
        self.title = json_data.get("title", "")
        self.description = json_data.get("description", "")

    @classmethod
    def from_json(cls, json_data, block_number, timestamp, position_in_the_block):
        instance = super().from_json(json_data, block_number, timestamp, position_in_the_block)
        instance.joining_fee = json_data.get("validatorJoiningFee", 8)
        instance.title = json_data.get("title", "")
        instance.description = json_data.get("description", "")
        return instance

    def to_json(self) -> dict:
        transaction = super().to_json()
        transaction["joiningFee"] = self.joining_fee
        transaction["title"] = self.title
        transaction["description"] = self.description
        return transaction


class ChangeVidaIdClaimingFeeProposalTransaction(Transaction):
    TYPE = "Change Vida Id Claiming Fee Proposal"

    def __init__(self, json_data: dict, block_number: int, timestamp: int, position_in_the_block: int,
                 claiming_fee: int = 8, title: str = "", description: str = ""):
        super().__init__(json_data, block_number, timestamp, position_in_the_block)
        self.claiming_fee = json_data.get("vidaIdClaimingFee", 8)
        self.title = json_data.get("title", "")
        self.description = json_data.get("description", "")

    @classmethod
    def from_json(cls, json_data, block_number, timestamp, position_in_the_block):
        instance = super().from_json(json_data, block_number, timestamp, position_in_the_block)
        instance.claiming_fee = json_data.get("vidaIdClaimingFee", 8)
        instance.title = json_data.get("title", "")
        instance.description = json_data.get("description", "")
        return instance

    def to_json(self) -> dict:
        transaction = super().to_json()
        transaction["claimingFee"] = self.claiming_fee
        transaction["title"] = self.title
        transaction["description"] = self.description
        return transaction


class ChangeVidaOwnerTxnFeeShareProposalTransaction(Transaction):
    TYPE = "Change VIDA Owner Txn Fee Share Proposal"

    def __init__(self, json_data: dict, block_number: int, timestamp: int, position_in_the_block: int,
                 fee_share: int = 8, title: str = "", description: str = ""):
        super().__init__(json_data, block_number, timestamp, position_in_the_block)
        self.fee_share = json_data.get("vidaOwnerTxnFeeShare", 8)
        self.title = json_data.get("title", "")
        self.description = json_data.get("description", "")

    @classmethod
    def from_json(cls, json_data, block_number, timestamp, position_in_the_block):
        instance = super().from_json(json_data, block_number, timestamp, position_in_the_block)
        instance.fee_share = json_data.get("vidaOwnerTxnFeeShare", 8)
        instance.title = json_data.get("title", "")
        instance.description = json_data.get("description", "")
        return instance

    def to_json(self) -> dict:
        transaction = super().to_json()
        transaction["feeShare"] = self.fee_share
        transaction["title"] = self.title
        transaction["description"] = self.description
        return transaction


class OtherProposalTransaction(Transaction):
    TYPE = "Other Proposal"

    def __init__(self, json_data: dict, block_number: int, timestamp: int, position_in_the_block: int,
                 title: str = "", description: str = ""):
        super().__init__(json_data, block_number, timestamp, position_in_the_block)
        self.title = json_data.get("title", "")
        self.description = json_data.get("description", "")

    @classmethod
    def from_json(cls, json_data, block_number, timestamp, position_in_the_block):
        instance = super().from_json(json_data, block_number, timestamp, position_in_the_block)
        instance.title = json_data.get("title", "")
        instance.description = json_data.get("description", "")
        return instance

    def to_json(self) -> dict:
        transaction = super().to_json()
        transaction["title"] = self.title
        transaction["description"] = self.description
        return transaction


class VoteOnProposalTransaction(Transaction):
    TYPE = "Vote On Proposal"

    def __init__(self, json_data: dict, block_number: int, timestamp: int, position_in_the_block: int,
                 proposal_hash: str = "0x", vote: int = 0, proposal_status: str = "ongoing"):
        super().__init__(json_data, block_number, timestamp, position_in_the_block)
        self.proposal_hash = json_data.get("proposalHash", "0x")
        self.vote = json_data.get("vote", 0)
        self.proposal_status = json_data.get("proposalStatus", "ongoing")

    @classmethod
    def from_json(cls, json_data, block_number, timestamp, position_in_the_block):
        instance = super().from_json(json_data, block_number, timestamp, position_in_the_block)
        instance.proposal_hash = json_data.get("proposalHash", "0x")
        instance.vote = json_data.get("vote", 0)
        instance.proposal_status = json_data.get("proposalStatus", "ongoing")
        return instance

    def to_json(self) -> dict:
        transaction = super().to_json()
        transaction["proposalHash"] = self.proposal_hash
        transaction["vote"] = self.vote
        return transaction


class TransactionResponse:
    def __init__(self, 
                 identifier: int = 0,
                 paid_total_fee: int = 0,
                 amount: int = 0,
                 paid_action_fee: int = 0,
                 nonce: int = 0,
                 transaction_hash: str = "0x",
                 time_stamp: int = 0,
                 fee_per_byte: int = 0,
                 size: int = 0,
                 sender: str = "0x",
                 success: bool = False,
                 block_number: int = 0,
                 position_in_the_block: int = 0,
                 vida_id: int = 0,
                 receiver: str = "0x",
                 data: bytes = b''):
        self.identifier = identifier
        self.paid_total_fee = paid_total_fee
        self.amount = amount
        self.paid_action_fee = paid_action_fee
        self.nonce = nonce
        self.transaction_hash = transaction_hash
        self.time_stamp = time_stamp
        self.fee_per_byte = fee_per_byte
        self.size = size
        self.sender = sender
        self.success = success
        self.block_number = block_number
        self.position_in_the_block = position_in_the_block
        self.vida_id = vida_id
        self.receiver = receiver
        self.data = data

    @classmethod
    def from_json(cls, json_data: dict):
        return cls(
            identifier=json_data.get('identifier', 0),
            paid_total_fee=json_data.get('paidTotalFee', 0),
            amount=json_data.get('amount', 0),
            paid_action_fee=json_data.get('paidActionFee', 0),
            nonce=json_data.get('nonce', 0),
            transaction_hash=json_data.get('transactionHash', '0x'),
            time_stamp=json_data.get('timeStamp', 0),
            fee_per_byte=json_data.get('feePerByte', 0),
            size=json_data.get('size', 0),
            sender=json_data.get('sender', '0x'),
            success=json_data.get('success', False),
            block_number=json_data.get('blockNumber', 0),
            position_in_the_block=json_data.get('positionInTheBlock', 0),
            vida_id=json_data.get('vidaId', 0),
            receiver=json_data.get('receiver', '0x'),
            data=bytes.fromhex(json_data.get('data', ''))
        )

    def to_json(self) -> dict:
        return {
            'identifier': self.identifier,
            'paidTotalFee': self.paid_total_fee,
            'amount': self.amount,
            'paidActionFee': self.paid_action_fee,
            'nonce': self.nonce,
            'transactionHash': self.transaction_hash,
            'timeStamp': self.time_stamp,
            'feePerByte': self.fee_per_byte,
            'size': self.size,
            'sender': self.sender,
            'success': self.success,
            'blockNumber': self.block_number,
            'positionInTheBlock': self.position_in_the_block,
            'vidaId': self.vida_id,
            'receiver': self.receiver,
            'data': self.data.hex()
        }

    def __str__(self) -> str:
        """Returns a string representation of the transaction."""
        return f"TransactionResponse(hash={self.transaction_hash}, sender={self.sender}, receiver={self.receiver}, amount={self.amount}, success={self.success})"

    def __repr__(self) -> str:
        """Returns a detailed string representation of the transaction."""
        return f"TransactionResponse(identifier={self.identifier}, hash={self.transaction_hash}, sender={self.sender}, receiver={self.receiver}, amount={self.amount}, success={self.success}, block={self.block_number}, position={self.position_in_the_block})"



import requests
from pwrpy.models.Delegator import Delegator


class Validator:
    def __init__(self, address, ip, bad_actor, voting_power, shares, delegators_count, status):
        self._address = address
        self._ip = ip
        self._bad_actor = bad_actor
        self._voting_power = voting_power
        self._shares = shares
        self._delegators_count = delegators_count
        self._delegators = []
        self._status = status

    @classmethod
    def from_json(cls, json_data):
        address = json_data.get("address", None)
        ip = json_data.get("ip", None)
        bad_actor = json_data.get("isBadActor", False)
        voting_power = json_data.get("votingPower", 0)
        shares = json_data.get("shares", 0)
        delegators_count = json_data.get("delegatorsCount", 0)
        status = json_data.get("status", None)
        return cls(address, ip, bad_actor, voting_power, shares, delegators_count, status)

    @property
    def address(self):
        return self._address

    @property
    def ip(self):
        return self._ip

    @property
    def bad_actor(self):
        return self._bad_actor

    @property
    def voting_power(self):
        return self._voting_power

    @property
    def shares(self):
        return self._shares

    @property
    def delegators_count(self):
        return self._delegators_count

    @property
    def status(self):
        return self._status

    def get_delegators(self, rpc_node_url):
        try:
            response = requests.get(
                rpc_node_url + "/validator/delegatorsOfValidator/?validatorAddress=" + self._address)

            # Check if the response was successful
            if response.status_code == 200:
                data = response.json()

                delegators_data = data['delegators']
                delegators_list = []

                for delegator_address, shares in delegators_data.items():
                    delegated_pwr = (shares * self._voting_power)
                    delegator = Delegator(
                        "0x" + delegator_address, self._address, shares, delegated_pwr)
                    delegators_list.append(delegator)

                return delegators_list
            elif response.status_code == 400:
                # If the response was a client error, raise an exception
                data = response.json()
                raise RuntimeError(f"Failed with HTTP error 400 and message: {data['message']}")
            else:
                # If the response was another kind of error, raise an exception
                raise RuntimeError(f"Failed with HTTP error code: {response.status_code}")

        except requests.HTTPError as http_err:
            raise RuntimeError(f"HTTP error occurred: {http_err}")
        except Exception as err:
            raise RuntimeError(f"An error occurred: {err}")



import struct
import time
from typing import List
from eth_utils import to_bytes


class TransactionBuilder:
    @staticmethod
    def get_transaction_base(identifier, nonce, chain_id, address, fee_per_byte):
        if len(address) == 42:
            address = address[2:]

        buffer = bytearray()
        buffer.extend(identifier.to_bytes(4, 'big'))
        buffer.append(int.from_bytes(chain_id, byteorder='big'))
        buffer.extend(nonce.to_bytes(4, 'big'))
        buffer.extend(fee_per_byte.to_bytes(8, 'big'))
        buffer.extend(bytes.fromhex(address))
        return bytes(buffer)

    @staticmethod
    def asset_address_validity(address: str):
        if address is None or (len(address) != 40 and len(address) != 42):
            raise RuntimeError("Invalid address")
    
    # Falcon transaction bytes
    @staticmethod
    def get_set_public_key_transaction(public_key, nonce, chain_id, address, fee_per_byte):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1001, nonce, chain_id, address, fee_per_byte)

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(len(public_key).to_bytes(2, byteorder='big'))
        buffer.extend(public_key)
        return bytes(buffer)
    
    @staticmethod
    def get_join_as_validator_transaction(ip: str, nonce, chain_id, address, fee_per_byte):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1002, nonce, chain_id, address, fee_per_byte)
        ip_bytes = ip.encode("utf-8")
    
        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(len(ip_bytes).to_bytes(2, byteorder='big'))
        buffer.extend(ip_bytes)
        return bytes(buffer)
    
    @staticmethod
    def get_delegate_transaction(validator: str, pwr_amount, nonce, chain_id, address, fee_per_byte):
        TransactionBuilder.asset_address_validity(validator)
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")
        
        if len(validator) == 42:
            validator = validator[2:]

        txn_base = TransactionBuilder.get_transaction_base(1003, nonce, chain_id, address, fee_per_byte)
    
        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(bytes.fromhex(validator))
        buffer.extend(pwr_amount.to_bytes(8, byteorder='big'))
        return bytes(buffer)
    
    @staticmethod
    def get_change_ip_transaction(new_ip: str, nonce, chain_id, address, fee_per_byte):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1004, nonce, chain_id, address, fee_per_byte)
        new_ip_bytes = new_ip.encode("utf-8")
    
        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(len(new_ip_bytes).to_bytes(2, byteorder='big'))
        buffer.extend(new_ip_bytes)
        return bytes(buffer)
    
    @staticmethod
    def get_claim_active_node_spot_transaction(nonce, chain_id, address, fee_per_byte):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1005, nonce, chain_id, address, fee_per_byte)
    
        buffer = bytearray()
        buffer.extend(txn_base)
        return bytes(buffer)
    
    @staticmethod
    def get_transfer_pwr_transaction(to: str, amount, nonce, chain_id, address, fee_per_byte):
        TransactionBuilder.asset_address_validity(to)
        if amount < 0:
            raise ValueError("Amount cannot be negative")
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        if len(to) == 42:
            to = to[2:]

        txn_base = TransactionBuilder.get_transaction_base(1006, nonce, chain_id, address, fee_per_byte)

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(bytes.fromhex(to))
        buffer.extend(amount.to_bytes(8, byteorder='big'))
        return bytes(buffer)
    
    # Governance Proposal Transactions
    @staticmethod
    def get_change_early_withdraw_penalty_proposal_transaction(title: str, description: str, early_withdrawal_time: int, 
                                                                    withdrawal_penalty: int, nonce: int, chain_id: int, 
                                                                    address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1009, nonce, chain_id, address, fee_per_byte)
        title_bytes = title.encode('utf-8')
        description_bytes = description.encode('utf-8')

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(len(title_bytes).to_bytes(4, byteorder='big'))
        buffer.extend(title_bytes)
        buffer.extend(early_withdrawal_time.to_bytes(8, byteorder='big'))
        buffer.extend(withdrawal_penalty.to_bytes(4, byteorder='big'))
        buffer.extend(description_bytes)
        return bytes(buffer)

    @staticmethod
    def get_change_fee_per_byte_proposal_transaction(title: str, description: str, new_fee_per_byte: int, 
                                                          nonce: int, chain_id: int, address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1010, nonce, chain_id, address, fee_per_byte)
        title_bytes = title.encode('utf-8')
        description_bytes = description.encode('utf-8')

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(len(title_bytes).to_bytes(4, byteorder='big'))
        buffer.extend(title_bytes)
        buffer.extend(new_fee_per_byte.to_bytes(8, byteorder='big'))
        buffer.extend(description_bytes)
        return bytes(buffer)

    @staticmethod
    def get_change_max_block_size_proposal_transaction(title: str, description: str, max_block_size: int, 
                                                            nonce: int, chain_id: int, address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1011, nonce, chain_id, address, fee_per_byte)
        title_bytes = title.encode('utf-8')
        description_bytes = description.encode('utf-8')

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(len(title_bytes).to_bytes(4, byteorder='big'))
        buffer.extend(title_bytes)
        buffer.extend(max_block_size.to_bytes(4, byteorder='big'))
        buffer.extend(description_bytes)
        return bytes(buffer)

    @staticmethod
    def get_change_max_txn_size_proposal_transaction(title: str, description: str, max_txn_size: int, 
                                                          nonce: int, chain_id: int, address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1012, nonce, chain_id, address, fee_per_byte)
        title_bytes = title.encode('utf-8')
        description_bytes = description.encode('utf-8')

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(len(title_bytes).to_bytes(4, byteorder='big'))
        buffer.extend(title_bytes)
        buffer.extend(max_txn_size.to_bytes(4, byteorder='big'))
        buffer.extend(description_bytes)
        return bytes(buffer)

    @staticmethod
    def get_change_overall_burn_percentage_proposal_transaction(title: str, description: str, burn_percentage: int, 
                                                                    nonce: int, chain_id: int, address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1013, nonce, chain_id, address, fee_per_byte)
        title_bytes = title.encode('utf-8')
        description_bytes = description.encode('utf-8')

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(len(title_bytes).to_bytes(4, byteorder='big'))
        buffer.extend(title_bytes)
        buffer.extend(burn_percentage.to_bytes(4, byteorder='big'))
        buffer.extend(description_bytes)
        return bytes(buffer)

    @staticmethod
    def get_change_reward_per_year_proposal_transaction(title: str, description: str, reward_per_year: int, 
                                                             nonce: int, chain_id: int, address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1014, nonce, chain_id, address, fee_per_byte)
        title_bytes = title.encode('utf-8')
        description_bytes = description.encode('utf-8')

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(len(title_bytes).to_bytes(4, byteorder='big'))
        buffer.extend(title_bytes)
        buffer.extend(reward_per_year.to_bytes(8, byteorder='big'))
        buffer.extend(description_bytes)
        return bytes(buffer)

    @staticmethod
    def get_change_validator_count_limit_proposal_transaction(title: str, description: str, validator_count_limit: int, 
                                                                   nonce: int, chain_id: int, address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1015, nonce, chain_id, address, fee_per_byte)
        title_bytes = title.encode('utf-8')
        description_bytes = description.encode('utf-8')

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(len(title_bytes).to_bytes(4, byteorder='big'))
        buffer.extend(title_bytes)
        buffer.extend(validator_count_limit.to_bytes(4, byteorder='big'))
        buffer.extend(description_bytes)
        return bytes(buffer)

    @staticmethod
    def get_change_validator_joining_fee_proposal_transaction(title: str, description: str, joining_fee: int, 
                                                                   nonce: int, chain_id: int, address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1016, nonce, chain_id, address, fee_per_byte)
        title_bytes = title.encode('utf-8')
        description_bytes = description.encode('utf-8')

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(len(title_bytes).to_bytes(4, byteorder='big'))
        buffer.extend(title_bytes)
        buffer.extend(joining_fee.to_bytes(8, byteorder='big'))
        buffer.extend(description_bytes)
        return bytes(buffer)

    @staticmethod
    def get_change_vida_id_claiming_fee_proposal_transaction(title: str, description: str, vida_id_claiming_fee: int, 
                                                                  nonce: int, chain_id: int, address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1017, nonce, chain_id, address, fee_per_byte)
        title_bytes = title.encode('utf-8')
        description_bytes = description.encode('utf-8')

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(len(title_bytes).to_bytes(4, byteorder='big'))
        buffer.extend(title_bytes)
        buffer.extend(vida_id_claiming_fee.to_bytes(8, byteorder='big'))
        buffer.extend(description_bytes)
        return bytes(buffer)

    @staticmethod
    def get_change_vida_owner_txn_fee_share_proposal_transaction(title: str, description: str, vida_owner_txn_fee_share: int, 
                                                                    nonce: int, chain_id: int, address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1018, nonce, chain_id, address, fee_per_byte)
        title_bytes = title.encode('utf-8')
        description_bytes = description.encode('utf-8')

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(len(title_bytes).to_bytes(4, byteorder='big'))
        buffer.extend(title_bytes)
        buffer.extend(vida_owner_txn_fee_share.to_bytes(4, byteorder='big'))
        buffer.extend(description_bytes)
        return bytes(buffer)

    @staticmethod
    def get_other_proposal_transaction(title: str, description: str, nonce: int, chain_id: int, 
                                            address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1019, nonce, chain_id, address, fee_per_byte)
        title_bytes = title.encode('utf-8')
        description_bytes = description.encode('utf-8')

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(len(title_bytes).to_bytes(4, byteorder='big'))
        buffer.extend(title_bytes)
        buffer.extend(description_bytes)
        return bytes(buffer)

    @staticmethod
    def get_vote_on_proposal_transaction(proposal_hash: str, vote: int, nonce: int, chain_id: int, 
                                              address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1020, nonce, chain_id, address, fee_per_byte)
        proposal_hash_bytes = bytes.fromhex(proposal_hash)

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(proposal_hash_bytes)
        buffer.extend(vote.to_bytes(1, byteorder='big'))
        return bytes(buffer)

    # Guardian Transactions
    @staticmethod
    def get_guardian_approval_transaction(wrapped_txns: List[bytes], nonce: int, chain_id: int, 
                                               address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1021, nonce, chain_id, address, fee_per_byte)

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(len(wrapped_txns).to_bytes(4, byteorder='big'))

        for txn in wrapped_txns:
            buffer.extend(len(txn).to_bytes(4, byteorder='big'))
            buffer.extend(txn)

        return bytes(buffer)

    @staticmethod
    def get_remove_guardian_transaction(nonce: int, chain_id: int, address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1022, nonce, chain_id, address, fee_per_byte)
        return txn_base

    @staticmethod
    def get_set_guardian_transaction(expiry_date: int, guardian_address: str, nonce: int, chain_id: int, 
                                          address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1023, nonce, chain_id, address, fee_per_byte)
        if len(guardian_address) == 42:
            guardian_address = guardian_address[2:]

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(expiry_date.to_bytes(8, byteorder='big'))
        buffer.extend(bytes.fromhex(guardian_address))
        return bytes(buffer)

    # Staking Transactions
    @staticmethod
    def get_move_stake_transaction(shares_amount: int, from_validator: str, to_validator: str, 
                                        nonce: int, chain_id: int, address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1024, nonce, chain_id, address, fee_per_byte)
        if len(from_validator) == 42:
            from_validator = from_validator[2:]
        if len(to_validator) == 42:
            to_validator = to_validator[2:]

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(shares_amount.to_bytes(8, byteorder='big'))
        buffer.extend(bytes.fromhex(from_validator))
        buffer.extend(bytes.fromhex(to_validator))
        return bytes(buffer)

    @staticmethod
    def get_remove_validator_transaction(validator_address: str, nonce: int, chain_id: int, 
                                              address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1025, nonce, chain_id, address, fee_per_byte)
        if len(validator_address) == 42:
            validator_address = validator_address[2:]

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(bytes.fromhex(validator_address))
        return bytes(buffer)

    @staticmethod
    def get_withdraw_transaction(shares_amount: int, validator: str, nonce: int, chain_id: int, 
                                      address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1026, nonce, chain_id, address, fee_per_byte)
        if len(validator) == 42:
            validator = validator[2:]

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(shares_amount.to_bytes(8, byteorder='big'))
        buffer.extend(bytes.fromhex(validator))
        return bytes(buffer)

    # VIDA Transactions
    @staticmethod
    def get_claim_vida_id_transaction(vida_id: int, nonce: int, chain_id: int, address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1028, nonce, chain_id, address, fee_per_byte)

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(vida_id.to_bytes(8, byteorder='big'))
        return bytes(buffer)

    @staticmethod
    def get_conduit_approval_transaction(vida_id: int, wrapped_txns: List[bytes], nonce: int, chain_id: int, 
                                              address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1029, nonce, chain_id, address, fee_per_byte)

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(vida_id.to_bytes(8, byteorder='big'))
        buffer.extend(len(wrapped_txns).to_bytes(4, byteorder='big'))

        for txn in wrapped_txns:
            buffer.extend(len(txn).to_bytes(4, byteorder='big'))
            buffer.extend(txn)

        return bytes(buffer)

    @staticmethod
    def get_payable_vida_data_transaction(vida_id: int, data: bytes, value: int, nonce: int, chain_id: int, 
                                               address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1030, nonce, chain_id, address, fee_per_byte)

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(vida_id.to_bytes(8, byteorder='big'))
        buffer.extend(len(data).to_bytes(4, byteorder='big'))
        buffer.extend(data)
        buffer.extend(value.to_bytes(8, byteorder='big'))
        return bytes(buffer)

    @staticmethod
    def get_remove_conduits_transaction(vida_id: int, conduits: List[str], nonce: int, chain_id: int, 
                                             address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1031, nonce, chain_id, address, fee_per_byte)

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(vida_id.to_bytes(8, byteorder='big'))

        for conduit in conduits:
            if len(conduit) == 42:
                conduit = conduit[2:]
            buffer.extend(bytes.fromhex(conduit))

        return bytes(buffer)

    @staticmethod
    def get_set_conduit_mode_transaction(vida_id: int, mode: int, conduit_threshold: int, conduits: List[str], 
                                              nonce: int, chain_id: int, address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1033, nonce, chain_id, address, fee_per_byte)

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(vida_id.to_bytes(8, byteorder='big'))
        buffer.extend(mode.to_bytes(1, byteorder='big'))
        buffer.extend(conduit_threshold.to_bytes(4, byteorder='big'))

        if conduits:
            buffer.extend(len(conduits).to_bytes(4, byteorder='big'))
            for conduit in conduits:
                if len(conduit) == 42:
                    conduit = conduit[2:]
                buffer.extend(bytes.fromhex(conduit))
        else:
            buffer.extend((0).to_bytes(4, byteorder='big'))

        return bytes(buffer)

    @staticmethod
    def get_set_vida_private_state_transaction(vida_id: int, private_state: bool, nonce: int, chain_id: int, 
                                                    address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1034, nonce, chain_id, address, fee_per_byte)

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(vida_id.to_bytes(8, byteorder='big'))
        buffer.extend((1 if private_state else 0).to_bytes(1, byteorder='big'))
        return bytes(buffer)

    @staticmethod
    def get_set_vida_to_absolute_public_transaction(vida_id: int, nonce: int, chain_id: int, 
                                                         address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1035, nonce, chain_id, address, fee_per_byte)

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(vida_id.to_bytes(8, byteorder='big'))
        return bytes(buffer)

    @staticmethod
    def get_add_vida_sponsored_addresses_transaction(vida_id: int, sponsored_addresses: List[str], 
                                                          nonce: int, chain_id: int, address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1036, nonce, chain_id, address, fee_per_byte)

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(vida_id.to_bytes(8, byteorder='big'))

        for address in sponsored_addresses:
            if len(address) == 42:
                address = address[2:]
            buffer.extend(bytes.fromhex(address))

        return bytes(buffer)

    @staticmethod
    def get_add_vida_allowed_senders_transaction(vida_id: int, allowed_senders: List[str], 
                                                      nonce: int, chain_id: int, address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1037, nonce, chain_id, address, fee_per_byte)

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(vida_id.to_bytes(8, byteorder='big'))

        for sender in allowed_senders:
            if len(sender) == 42:
                sender = sender[2:]
            buffer.extend(bytes.fromhex(sender))

        return bytes(buffer)

    @staticmethod
    def get_remove_vida_allowed_senders_transaction(vida_id: int, allowed_senders: List[str], 
                                                         nonce: int, chain_id: int, address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1038, nonce, chain_id, address, fee_per_byte)

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(vida_id.to_bytes(8, byteorder='big'))

        for sender in allowed_senders:
            if len(sender) == 42:
                sender = sender[2:]
            buffer.extend(bytes.fromhex(sender))

        return bytes(buffer)

    @staticmethod
    def get_remove_sponsored_addresses_transaction(vida_id: int, sponsored_addresses: List[str], 
                                                        nonce: int, chain_id: int, address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1039, nonce, chain_id, address, fee_per_byte)

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(vida_id.to_bytes(8, byteorder='big'))

        for address in sponsored_addresses:
            if len(address) == 42:
                address = address[2:]
            buffer.extend(bytes.fromhex(address))

        return bytes(buffer)

    @staticmethod
    def get_set_pwr_transfer_rights_transaction(vida_id: int, owner_can_transfer_pwr: bool, 
                                                     nonce: int, chain_id: int, address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1040, nonce, chain_id, address, fee_per_byte)

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(vida_id.to_bytes(8, byteorder='big'))
        buffer.extend((1 if owner_can_transfer_pwr else 0).to_bytes(1, byteorder='big'))
        return bytes(buffer)

    @staticmethod
    def get_transfer_pwr_from_vida_transaction(vida_id: int, receiver: str, amount: int, 
                                                    nonce: int, chain_id: int, address: str, fee_per_byte: int):
        if nonce < 0:
            raise ValueError("Nonce cannot be negative")

        txn_base = TransactionBuilder.get_transaction_base(1041, nonce, chain_id, address, fee_per_byte)
        if len(receiver) == 42:
            receiver = receiver[2:]

        buffer = bytearray()
        buffer.extend(txn_base)
        buffer.extend(vida_id.to_bytes(8, byteorder='big'))
        buffer.extend(bytes.fromhex(receiver))
        buffer.extend(amount.to_bytes(8, byteorder='big'))
        return bytes(buffer)
    
